################################################################################
############################# INNER FUNCTIONS ##################################
################################################################################

#' Compute details about agreement in number of snow covered days
#'
#' \code{snow.days.comparison} is an internal function of
#' \code{\link{plt.snow.days.comparison()}}. It compares the two given datasets.
#' For details see \code{\link{plt.snow.days.comparison()}}
#'
#' @param data.1 The training dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.2 The dataset to be compared to training dataset (a vectdata object,
#' see \code{\link{vectorize.list()}})
#' @param swe.threshold The threshold SWE value to define snow days
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#'
#'@return A truth table (one column per year) comparing the two datasets.
#' Lines are 0 = true negative, 1 = false negative, 2 = false positive,
#' 4 =  true positive.
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
snow.days.comparison <- function(data.1, data.2, ncores, swe.threshold = 0.005){

  years <- lubridate::year(data.1$time.date)
  years <- ifelse(lubridate::month(data.1$time.date) %in% c(9, 10, 11, 12),
                  years+1, years)

  if(!identical(dim(data.1$data), dim(data.2$data))) {
    stop("data.1 and data.2 should be of identitacale shape")
  }

  indices=list()
  for(yr in unique(years)){
    indices[[paste(yr)]]=which(years==yr)
  }
  snow.days <- snowDaysComparisonCpp(data.1$data,
                                     data.2$data,
                                     indices,
                                     ncores,
                                     swe.threshold)
  return(snow.days)
}

# get.ratio <- function(data,col1,col2) {
#   data.out <- unlist(lapply(data,function(x) {x[col1,]/(x[col1,]+x[col2,])}))
#   data.out[is.nan(data.out)] <- 0
#   data.out[is.infinite(data.out)] <- 1
#   data.out[data.out >1 ] <- 1
#   return(mean(data.out))
# }

get.map <- function(data.out, snow.days) {
  get.truth.table.map <- function(snow.days){
    snow.days <- lapply(snow.days,rowSums)
    snow.days <- lapply(snow.days,function(x){x/sum(x)})
    snow.days <- matrix(unlist(snow.days),ncol=4,byrow = TRUE)
    snow.days <- lapply(seq_len(ncol(snow.days)), function(i) matrix(snow.days[,i]))
    names(snow.days)=c("true negative","false negative","false positive", "true positive")
    return(snow.days)
  }
  data.out <- append(data.out, get.truth.table.map(snow.days))
  class(data.out) <- c("vectdata", "list")
  data.out$data <- NULL
  return(data.out)
}

get.plot.data <- function(data) {
  data <- Reduce(`+`, data)
  data <- cbind(data,total=rowSums(data))
  data <- sweep(data, 2, colSums(data), FUN="/")[c(4,1,2,3),]
}

get.metrics <- function(data) {
  out = list()
  out$false_negative <- mean(unlist(lapply(data,function(x){return(mean(x[2,]))})))
  out$false_positive <- mean(unlist(lapply(data,function(x){return(mean(x[3,]))})))
  return(out)
}

cut.season <- function(data, season) {
  seasons.months <- list("Winter"=c(12,1,2),"Spring"=c(3,4,5),
                         "Summer"=c(6,7,8),"Fall"=c(9,10,11))
  sel <- which(lubridate::month(data$time.date) %in% seasons.months[[season]])
  data$time <- data$time[sel]
  data$time.date <- data$time.date[sel]
  data$data <- data$data[,sel]
  return(data)
}


################################################################################
############################# SINGLE MODEL RUN #################################
################################################################################


#' Plot details about agreement in number of snow covered days
#'
#' \code{plt.snow.days.comparison} produces barplots (see
#' \code{\link{barplot()}}) of the fraction of agreement and disagreement in term
#' of snow covered days between the training and the model dataset and between
#' the training and corrected dataset. Agreement is snow per year, aggregated for
#' all the pixels. Agreement is snow in term of true positive (both predict a
#' snow covered day), true negative (both predict a snow free day),
#' false positive (training predicts a snow free day, but model/corrected
#' predicts a snow covered day), and false negative (training predicts a snow
#' covered day, but model/corrected predicts a snow free day). Data to
#' produce the plots are also saved into a RDS file
#'
#' @param data.training The training dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.model The model dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.corrected the corrected dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param out.plot.dir Path in which the RDS files should be saved. Three files are
#' created: snow_season_training, snow_season_model, snow_season_corrected
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param swe.threshold The threshold SWE value to define snow days
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
plt.snow.days.comparison<- function(data.training, data.model,
                                    data.corrected, ncores, swe.threshold) {

  out = list()

  #Compute snow days table
  snow.days.model <- snow.days.comparison(data.training,
                                                   data.model,
                                                   ncores, swe.threshold)

  snow.days.corrected <- snow.days.comparison(data.training,
                                                    data.corrected,
                                                    ncores, swe.threshold)

  #Normalize truth table
  snow.days.model <- lapply(snow.days.model,
                               function(x) {sweep(x, 2, colSums(x), FUN="/")})
  snow.days.corrected <- lapply(snow.days.corrected,
                               function(x) {sweep(x, 2, colSums(x), FUN="/")})

  #Get maps of truth table
  out$data.model <- get.map(data.model, snow.days.model)
  out$data.corrected <- get.map(data.training, snow.days.corrected)

  #Get metrics
  out$model <- get.metrics(snow.days.model)
  out$corrected <- get.metrics(snow.days.corrected)

  #Sum all pixels and add total column for bar plots
  plot.data.model <- get.plot.data(snow.days.model)
  plot.data.corrected <- get.plot.data(snow.days.corrected)

  barplot(plot.data.model, col = c(3, 4, 2, 7),
          space=c(0,rep(2.5,dim(plot.data.model)[2]-2),4),
          density=rep(rep(NA,4), rep(50,4)), ylim=c(0,1.15),
          main ="Snow days comparison", ylab="Fraction")

  barplot(plot.data.corrected,col=c(3,4,2,7), yaxt="n",
          space=c(1.4,rep(2.5,dim(plot.data.corrected)[2]-2),4),add=TRUE,xaxt="n")
  legend("top", legend=c("model","corrected", "true positive", "true negative",
                         "false negative", "false positive"), fill=c(1,1,3,4,2,7),
         density = c(NA,50,1000,NA,NA,NA), bty="n", ncol=6)

  return(out)
}



#' XXXXXXXXX
#'
#' \code{snow.days.comparison.calib.seasonnal} produces barplots (see
#' \code{\link{barplot()}}) of the fraction of agreement and disagreement in term
#' of snow covered days between the training and the model dataset and between
#' the training and corrected dataset. Agreement is snow per year, aggregated for
#' all the pixels. Agreement is snow in term of true positive (both predict a
#' snow covered day), true negative (both predict a snow free day),
#' false positive (training predicts a snow free day, but model/corrected
#' predicts a snow covered day), and false negative (training predicts a snow
#' covered day, but model/corrected predicts a snow free day). Data to
#' produce the plots are also saved into a RDS file
#'
#' @param data.training The training dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.model The model dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.corrected the corrected dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param out.plot.dir Path in which the RDS files should be saved. Three files are
#' created: snow_season_training, snow_season_model, snow_season_corrected
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param swe.threshold The threshold SWE value to define snow days
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
snow.days.comparison.seasonnal <- function(data.training, data.model,
                                                 data.corrected, ncores, swe.threshold) {

  seasons <- c("Winter","Spring","Summer","Fall")

  out=list()

  for(season in seasons){
    #Trim data to season
    training <- cut.season(data.training, season)
    model <- cut.season(data.model, season)
    corrected <- cut.season(data.corrected, season)

    #Compute truth table
    snow.days.model <- snow.days.comparison(training, model,
                                               ncores, swe.threshold)
    snow.days.corrected <- snow.days.comparison(training, corrected,
                                                ncores, swe.threshold)

    #Normalize truth table
    snow.days.model <- lapply(snow.days.model, function(x) {sweep(x, 2, colSums(x), FUN="/")})
    snow.days.corrected <- lapply(snow.days.corrected, function(x) {sweep(x, 2, colSums(x), FUN="/")})

    #Get maps
    out$season.model[[season]] <- get.map(training, snow.days.model)
    out$season.corrected[[season]] <- get.map(corrected, snow.days.corrected)

    #Get metrics
    out[[season]]$model <- get.metrics(snow.days.model)
    out[[season]]$corrected <- get.metrics(snow.days.corrected)
  }
  return(out)
}

#' XXXXX
#'
#' \code{plt.snow.days.comparison} XXX
#' \code{\link{barplot()}}) XXXX
#'
#' @param data.training The training dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.model The model dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.corrected the corrected dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param out.plot.dir Path in which the RDS files should be saved. Three files are
#' created: snow_season_training, snow_season_model, snow_season_corrected
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param swe.threshold The threshold SWE value to define snow days
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
plt.snow.days.maps <- function(data.model, data.corrected,
                               data.model.seasons, data.corrected.seasons,
                               out.plot.dir) {

  pdf(paste0(out.plot.dir,"/snow_days_maps.pdf"),width=8,height=8)
  par(mar=c(3,3,2,5), mgp=c(1.7,0.5,0),mfrow=c(4,2))

  for(var in c("true positive", "true negative", "false negative","false positive")) {
    plot(data.model, var = var, zlim=c(0,1), main = paste("model calib", var))
    plot(data.corrected, var = var, zlim=c(0,1), main = paste("Corrected calib", var))
  }

  par(mar=c(3,3,2,5), mgp=c(1.7,0.5,0),mfrow=c(4,2), oma = c(0,0,2,0))
  seasons <- c("Winter","Spring","Summer","Fall")
  for(season in seasons) {
    for(var in c("true positive", "true negative", "false negative","false positive")) {
      plot(data.model.seasons[[season]], var = var, zlim = c(0,1),
           main=paste("model calib", var))
      plot(data.corrected.seasons[[season]], var = var, zlim = c(0,1),
           main=paste("Corrected calib", var))
    }
    title(season, outer = TRUE)
  }
  dev.off()
}

################################################################################
####################### CALIBRATION AND VALIDATION #############################
################################################################################

#' Plot details about agreement in number of snow covered days
#'
#' \code{plt.snow.days.comparison} produces barplots (see
#' \code{\link{barplot()}}) of the fraction of agreement and disagreement in term
#' of snow covered days between the training and the model dataset and between
#' the training and corrected dataset. Agreement is snow per year, aggregated for
#' all the pixels. Agreement is snow in term of true positive (both predict a
#' snow covered day), true negative (both predict a snow free day),
#' false positive (training predicts a snow free day, but model/corrected
#' predicts a snow covered day), and false negative (training predicts a snow
#' covered day, but model/corrected predicts a snow free day). Data to
#' produce the plots are also saved into a RDS file
#'
#' @param data.training The training dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.model The model dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.corrected the corrected dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param out.plot.dir Path in which the RDS files should be saved. Three files are
#' created: snow_season_training, snow_season_model, snow_season_corrected
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param swe.threshold The threshold SWE value to define snow days
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
plt.snow.days.comparison.calib <- function(data.training.calib, data.model.calib,
                                           data.corrected.calib, data.training.valid,
                                           data.model.valid, data.corrected.valid,
                                           ncores, swe.threshold) {

  out = list()

  #Compute snow days table
  snow.days.model.calib <- snow.days.comparison(data.training.calib,
                                                   data.model.calib,
                                                   ncores, swe.threshold)

  snow.days.corrected.calib <- snow.days.comparison(data.training.calib,
                                                    data.corrected.calib,
                                                    ncores, swe.threshold)

  snow.days.model.valid <- snow.days.comparison(data.training.valid,
                                                   data.model.valid,
                                                   ncores, swe.threshold)

  snow.days.corrected.valid <- snow.days.comparison(data.training.valid,
                                                    data.corrected.valid,
                                                    ncores, swe.threshold)

  #Normalize truth table
  snow.days.model.calib <- lapply(snow.days.model.calib,
                                     function(x) {sweep(x, 2, colSums(x), FUN="/")})
  snow.days.corrected.calib <- lapply(snow.days.corrected.calib,
                                      function(x) {sweep(x, 2, colSums(x), FUN="/")})
  snow.days.model.valid <- lapply(snow.days.model.valid,
                                     function(x) {sweep(x, 2, colSums(x), FUN="/")})
  snow.days.corrected.valid <- lapply(snow.days.corrected.valid,
                                      function(x) {sweep(x, 2, colSums(x), FUN="/")})

  #Get maps of truth table
  out$data.model.calib <- get.map(data.model.calib, snow.days.model.calib)
  out$data.corrected.calib <- get.map(data.training.calib, snow.days.corrected.calib)
  out$data.model.valid <- get.map(data.model.valid, snow.days.model.valid)
  out$data.corrected.valid <- get.map(data.training.valid, snow.days.corrected.valid)


  #Get metrics
  out$calib$model <- get.metrics(snow.days.model.calib)
  out$calib$corrected <- get.metrics(snow.days.corrected.calib)
  out$valid$model <- get.metrics(snow.days.model.valid)
  out$valid$corrected <- get.metrics(snow.days.corrected.valid)

  #Sum all pixels and add total column for bar plots
  plot.data.model.calib <- get.plot.data(snow.days.model.calib)
  plot.data.corrected.calib <- get.plot.data(snow.days.corrected.calib)
  plot.data.model.valid <- get.plot.data(snow.days.model.valid)
  plot.data.corrected.valid <- get.plot.data(snow.days.corrected.valid)

  barplot(plot.data.model.calib, col = c(3, 4, 2, 7),
          space=c(0,rep(2.5,dim(plot.data.model.calib)[2]-2),4),
          density=rep(rep(NA,4), rep(50,4)), ylim=c(0,1.15),
          main ="Snow days comparison calibration", ylab="Fraction")

  barplot(plot.data.corrected.calib,col=c(3,4,2,7), yaxt="n",
          space=c(1.4,rep(2.5,dim(plot.data.corrected.calib)[2]-2),4),add=TRUE,xaxt="n")
  legend("top", legend=c("model","corrected", "true positive", "true negative",
                         "false negative", "false positive"), fill=c(1,1,3,4,2,7),
         density = c(NA,50,1000,NA,NA,NA), bty="n", ncol=6)

  barplot(plot.data.model.valid, col = c(3, 4, 2, 7),
          space=c(0,rep(2.5,dim(plot.data.model.valid)[2]-2),4),
          density=rep(rep(NA,4), rep(50,4)), ylim=c(0,1.15),
          main =  "Snow days comparison validation", ylab="Fraction")
  barplot(plot.data.corrected.valid,col=c(3,4,2,7), yaxt="n",
          space=c(1.4,rep(2.5,dim(plot.data.corrected.valid)[2]-2),4),add=TRUE,xaxt="n")
  legend("top", legend=c("model","corrected", "true positive", "true negative",
                         "false negative", "false positive"), fill=c(1,1,3,4,2,7),
         density = c(NA,50,1000,NA,NA,NA), bty="n", ncol=6)

  return(out)
}



#' XXXXXXXXX
#'
#' \code{snow.days.comparison.calib.seasonnal} produces barplots (see
#' \code{\link{barplot()}}) of the fraction of agreement and disagreement in term
#' of snow covered days between the training and the model dataset and between
#' the training and corrected dataset. Agreement is snow per year, aggregated for
#' all the pixels. Agreement is snow in term of true positive (both predict a
#' snow covered day), true negative (both predict a snow free day),
#' false positive (training predicts a snow free day, but model/corrected
#' predicts a snow covered day), and false negative (training predicts a snow
#' covered day, but model/corrected predicts a snow free day). Data to
#' produce the plots are also saved into a RDS file
#'
#' @param data.training The training dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.model The model dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.corrected the corrected dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param out.plot.dir Path in which the RDS files should be saved. Three files are
#' created: snow_season_training, snow_season_model, snow_season_corrected
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param swe.threshold The threshold SWE value to define snow days
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
snow.days.comparison.seasonnal.calib <- function(data.training.calib, data.model.calib,
                                                 data.corrected.calib, data.training.valid,
                                                 data.model.valid, data.corrected.valid,
                                                 ncores, swe.threshold) {

  seasons <- c("Winter","Spring","Summer","Fall")

  out=list()

  for(season in seasons){
    #Trim data to season
    training.calib <- cut.season(data.training.calib, season)
    model.calib <- cut.season(data.model.calib, season)
    corrected.calib <- cut.season(data.corrected.calib, season)
    training.valid <- cut.season(data.training.valid, season)
    model.valid <- cut.season(data.model.valid, season)
    corrected.valid <- cut.season(data.corrected.valid, season)

    #Compute truth table
    snow.days.model.calib <- snow.days.comparison(training.calib,
                                                     model.calib,
                                                     ncores,
                                                     swe.threshold)
    snow.days.corrected.calib <- snow.days.comparison(training.calib,
                                                      corrected.calib,
                                                      ncores,
                                                      swe.threshold)
    snow.days.model.valid <- snow.days.comparison(training.valid,
                                                     model.valid,
                                                     ncores,
                                                     swe.threshold)
    snow.days.corrected.valid <- snow.days.comparison(training.valid,
                                                      corrected.valid,
                                                      ncores,
                                                      swe.threshold)

    #Normalize truth table
    snow.days.model.calib <- lapply(snow.days.model.calib, function(x) {sweep(x, 2, colSums(x), FUN="/")})
    snow.days.corrected.calib <- lapply(snow.days.corrected.calib, function(x) {sweep(x, 2, colSums(x), FUN="/")})
    snow.days.model.valid <- lapply(snow.days.model.valid, function(x) {sweep(x, 2, colSums(x), FUN="/")})
    snow.days.corrected.valid <- lapply(snow.days.corrected.valid, function(x) {sweep(x, 2, colSums(x), FUN="/")})

    #Get maps
    out$season.model.calib[[season]] <- get.map(training.calib, snow.days.model.calib)
    out$season.corrected.calib[[season]] <- get.map(corrected.calib, snow.days.corrected.calib)
    out$season.model.valid[[season]] <- get.map(model.valid, snow.days.model.valid)
    out$season.corrected.valid[[season]] <- get.map(corrected.valid, snow.days.corrected.valid)

    #Get metrics
    out$calib$model[[season]] <- get.metrics(snow.days.model.calib)
    out$calib$corrected[[season]] <- get.metrics(snow.days.corrected.calib)
    out$valid$model[[season]] <- get.metrics(snow.days.model.valid)
    out$valid$corrected[[season]] <- get.metrics(snow.days.corrected.valid)
  }
  return(out)
}

#' XXXXX
#'
#' \code{plt.snow.days.comparison} XXX
#' \code{\link{barplot()}}) XXXX
#'
#' @param data.training The training dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.model The model dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.corrected the corrected dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param out.plot.dir Path in which the RDS files should be saved. Three files are
#' created: snow_season_training, snow_season_model, snow_season_corrected
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param swe.threshold The threshold SWE value to define snow days
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
plt.snow.days.maps.calib <- function(data.model.calib, data.corrected.calib,
                                     data.model.valid, data.corrected.valid,
                                     data.model.calib.seasons, data.corrected.calib.seasons,
                                     data.model.valid.seasons, data.corrected.valid.seasons,
                                     out.plot.dir) {

  pdf(paste0(out.plot.dir,"/snow_days_maps.pdf"),width=12,height=8)
  par(mar=c(3,3,2,5), mgp=c(1.7,0.5,0),mfrow=c(4,4))

  for(var in c("true positive", "true negative", "false negative","false positive")) {
    plot(data.model.calib, var = var, zlim=c(0,1), main = paste("model calib", var))
    plot(data.corrected.calib, var = var, zlim=c(0,1), main = paste("Corrected calib", var))
    plot(data.model.valid, var = var, zlim=c(0,1), main = paste("model valid", var))
    plot(data.corrected.valid, var = var, zlim=c(0,1), main = paste("Corrected valid", var))
  }

  par(mar=c(3,3,2,5), mgp=c(1.7,0.5,0),mfrow=c(4,4), oma = c(0,0,2,0))
  seasons <- c("Winter","Spring","Summer","Fall")
  for(season in seasons) {
    for(var in c("true positive", "true negative", "false negative","false positive")) {
      plot(data.model.calib.seasons[[season]], var = var, zlim = c(0,1),
           main=paste("model calib", var))
      plot(data.corrected.calib.seasons[[season]], var = var, zlim = c(0,1),
           main=paste("Corrected calib", var))
      plot(data.model.valid.seasons[[season]], var = var, zlim = c(0,1),
           main=paste("model valid", var))
      plot(data.corrected.valid.seasons[[season]], var = var, zlim = c(0,1),
           main=paste("Corrected valid", var))
    }
    title(season, outer = TRUE)
  }
  dev.off()
}
