################################################################################
############################# INNER FUNCTIONS ##################################
################################################################################

#' Compute the seasonal snow cover length
#'
#' \code{snow.season.length()} is an internal function called by
#' \code{\link{plt.snow.season.length()}}). It splits the work between
#' pixels and call \code{\link{snow.season.length.pixel()}}) on each pixel
#' timeseries.
#'
#' @param data The training dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param swe.threshold The threshold value to define snow day in order to
#' compute the length of the snow season
#' @param ma.value Number of day to apply for moving average function on SWE
#' before computing the length of the snow seasons
#'
#' @return A list containing for each pixel a vector containing the length of
#' the snow season for each individual year
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
snow.season.length <- function(data, ncores = 1, swe.threshold = 0.05,
                               ma.value = 7){
  #Compute indices list for each year
  years <- lubridate::year(data$time.date)
  years <- ifelse(lubridate::month(data$time.date) %in% c(9, 10, 11, 12),
                  years+1, years)
  indices=list()
  for(yr in unique(years)){
    indices[[paste(yr)]]=which(years==yr)
  }

  snow.season <- snowSeasonLengthCpp(data$data, indices, ncores, swe.threshold,
                                     ma.value)

  return(snow.season)
}


################################################################################
############################# SINGLE MODEL RUN #################################
################################################################################

#' Plot details about the seasonal snow cover length
#'
#' \code{plt.snow.season.length} produces violin plots (see
#' \code{\link{vioplot()}}) of the annual snow seasons length for each pixels.
#' Snow season lengths are shown for the training dataset, the model dataset,
#' and the corrected dataset.Snow seasons length are computed as the longest
#' sequence of days where SWE exceeds the threshold given as input for each year
#' (September to August). A circular moving average is applied to SWE time series
#' beforehand (controlled through the \code{ma.value} input parameter). Data to
#' produce the plots are also saved into a RDS file
#'
#' @param data.training The training dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.model The model dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.corrected the corrected dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param out.plot.dir Path in which the RDS files should be saved. Three files are
#' created: snow_season_training, snow_season_model, snow_season_corrected
#' @param swe.threshold The threshold value to define snow day in order to
#' compute the length of the snow season
#' @param ma.value Number of day to apply for moving average function on SWE
#' before computing the length of the snow seasons.
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
plt.snow.season.length <- function(data.training, data.model,
                                   data.corrected, ncores,
                                   swe.threshold, ma.value) {

  # Compute snow season length for each pixels
  snow.season.training <- snow.season.length(data.training, ncores,
                                                 swe.threshold, ma.value)
  snow.season.model <- snow.season.length(data.model, ncores,
                                                   swe.threshold, ma.value)
  snow.season.corrected <- snow.season.length(data.corrected, ncores,
                                                    swe.threshold, ma.value)

  #Plot absolute values
  calib.years <- colnames(snow.season.training)
  data.length <- length(calib.years)
# 
#   vioplot::vioplot(snow.season.training, at=seq(from=1,length.out=data.length, by=4),
#                    col=2, colMed =1, ylab="Snow season length (day)", main = "Snow season length",
#                    names = calib.years)
#   legend("topleft", c("training","model","corrected"), col=c(2,5,6), lwd=7,
#          bty="n", ncol=1)
#   vioplot::vioplot(snow.season.model, at = seq(from = 2,length.out = data.length, by = 4),
#                    col = 5, add = TRUE, colMed =1)
#   vioplot::vioplot(snow.season.corrected, at = seq(from = 3, length.out = data.length, by = 4),
#                    col = 6, add = TRUE, colMed =1)

  #Plot relative error calibration
  snow.season.model.rel <- (snow.season.model-snow.season.training)/snow.season.training
  snow.season.corrected.rel <- (snow.season.corrected-snow.season.training)/snow.season.training
  snow.season.model.rel[snow.season.model.rel>1]<-1
  snow.season.corrected.rel[snow.season.corrected.rel>1]<-1
  snow.season.model.rel[is.na(snow.season.model.rel)]<-0
  snow.season.corrected.rel[is.na(snow.season.corrected.rel)]<-0
# 
#   vioplot::vioplot(snow.season.model.rel, at=seq(from=1,length.out=data.length, by=4),
#                    col=5, colMed =1, ylab="Rel. error (-)", main = "Relative error",yalim=c(-1,1),
#                    names = calib.years, ylim=c(-1,1))
#   vioplot::vioplot(snow.season.corrected.rel, at = seq(from = 2, length.out = data.length, by = 4),
#                    col = 6, add = TRUE, colMed =1)
#   legend("topleft", c("model","corrected"), col=c(5,6), lwd=7,
#          bty="n", ncol=3)
#   legend("bottomleft", c(paste("model mean",
#                                round(mean(snow.season.model.rel),2)),
#                          paste("model mae",
#                                round(mean(abs(snow.season.model.rel)),2)),
#                          paste("corrected mean",
#                                round(mean(snow.season.corrected.rel),2)),
#                          paste("corrected mae",
#                                round(mean(abs(snow.season.corrected.rel)),2))),
#          bty="n", ncol=3)

  #Prepare output data
  out=list()
  out$model$mean.error <- mean(snow.season.model.rel)
  out$model$mae <- mean(abs(snow.season.model.rel))
  out$corrected$mean.error <- mean(snow.season.corrected.rel)
  out$corrected$mae <- mean(abs(snow.season.corrected.rel))

  #Add data to produce map plots from output
  data.training$snow.season <- matrix(rowMeans(snow.season.training))
  data.training$data <- NULL
  out$data.training <- data.training
  data.model$snow.season <- matrix(rowMeans(snow.season.model))
  data.model$data <- NULL
  out$data.model <- data.model
  data.corrected$snow.season <- matrix(rowMeans(snow.season.corrected))
  data.corrected$data <- NULL
  out$data.corrected <- data.corrected

  return(out)
}


plt.snow.season.maps <- function(snow.season.training, snow.season.model,
                                 snow.season.corrected, out.plot.dir) {

  pdf(paste0(out.plot.dir,"/snow_season_maps.pdf"),width=16,height=6)

  par(mar=c(3,4,2,7), mgp=c(1.7,0.5,0),mfrow=c(1,3))

  #Plot map of absolute values
  plot(snow.season.training, var="snow.season", zlim=c(0,366),
       main="training", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.model, var="snow.season", zlim=c(0,366),
       main="model", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.corrected, var="snow.season", zlim=c(0,366),
       main="Corrected", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)



  par(mar=c(3,4,2,8), mgp=c(1.7,0.5,0),mfrow=c(1,2))
  #Compute relative error
  snow.season.model$snow.season.rel <- (snow.season.model$snow.season -
                                                   snow.season.training$snow.season) /
                                                 snow.season.training$snow.season
  snow.season.corrected$snow.season.rel <- (snow.season.corrected$snow.season -
                                                    snow.season.training$snow.season) /
                                                  snow.season.training$snow.season
  snow.season.model$snow.season.rel[snow.season.model$snow.season.rel>1] <- 1
  snow.season.corrected$snow.season.rel[snow.season.corrected$snow.season.rel>1] <- 1


  #Plot relative error
  plot(snow.season.model, var="snow.season.rel", zlim=c(-1,1),
       main="model", legend.lab="Error (%)",
       legend.cex=0.9, legend.line=2.7)
  plot(snow.season.corrected, var="snow.season.rel", zlim=c(-1,1),
       main="Corrected", legend.lab="Error (%)",
       legend.cex=0.9, legend.line=2.7)
  dev.off()

}

################################################################################
####################### CALIBRATION AND VALIDATION #############################
################################################################################

#' Plot details about the seasonal snow cover length
#'
#' \code{plt.snow.season.length} produces violin plots (see
#' \code{\link{vioplot()}}) of the annual snow seasons length for each pixels.
#' Snow season lengths are shown for the training dataset, the model dataset,
#' and the corrected dataset.Snow seasons length are computed as the longest
#' sequence of days where SWE exceeds the threshold given as input for each year
#' (September to August). A circular moving average is applied to SWE time series
#' beforehand (controlled through the \code{ma.value} input parameter). Data to
#' produce the plots are also saved into a RDS file
#'
#' @param data.training The training dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.model The model dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.corrected the corrected dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param out.plot.dir Path in which the RDS files should be saved. Three files are
#' created: snow_season_training, snow_season_model, snow_season_corrected
#' @param swe.threshold The threshold value to define snow day in order to
#' compute the length of the snow season
#' @param ma.value Number of day to apply for moving average function on SWE
#' before computing the length of the snow seasons.
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
plt.snow.season.length.calib <- function(data.training.calib, data.model.calib,
                                         data.corrected.calib, data.training.valid,
                                         data.model.valid, data.corrected.valid,
                                         ncores, swe.threshold, ma.value) {

  # Compute snow season length for each pixels
  snow.season.training.calib <- snow.season.length(data.training.calib, ncores,
                                                 swe.threshold, ma.value)
  snow.season.model.calib <- snow.season.length(data.model.calib, ncores,
                                                   swe.threshold, ma.value)
  snow.season.corrected.calib <- snow.season.length(data.corrected.calib, ncores,
                                                    swe.threshold, ma.value)

  snow.season.training.valid <- snow.season.length(data.training.valid, ncores,
                                                 swe.threshold, ma.value)
  snow.season.model.valid <- snow.season.length(data.model.valid, ncores,
                                                   swe.threshold, ma.value)
  snow.season.corrected.valid <- snow.season.length(data.corrected.valid, ncores,
                                                    swe.threshold, ma.value)


  #Plot absolute values calibration
  calib.years <- colnames(snow.season.training.calib)
  data.length <- length(calib.years)
# 
#   vioplot::vioplot(snow.season.training.calib, at=seq(from=1,length.out=data.length, by=4),
#                    col=2, colMed =1, ylab="Snow season length (day)", main = "Calibration",
#                    names = calib.years)
#   legend("topleft", c("training","model","corrected"), col=c(2,5,6), lwd=7,
#          bty="n", ncol=1)
#   vioplot::vioplot(snow.season.model.calib, at = seq(from = 2,length.out = data.length, by = 4),
#                    col = 5, add = TRUE, colMed =1)
#   vioplot::vioplot(snow.season.corrected.calib, at = seq(from = 3, length.out = data.length, by = 4),
#                    col = 6, add = TRUE, colMed =1)

  #Plot relative error calibration
  snow.season.model.calib.rel <- (snow.season.model.calib-snow.season.training.calib)/snow.season.training.calib
  snow.season.corrected.calib.rel <- (snow.season.corrected.calib-snow.season.training.calib)/snow.season.training.calib
  snow.season.model.calib.rel[snow.season.model.calib.rel>1]<-1
  snow.season.corrected.calib.rel[snow.season.corrected.calib.rel>1]<-1
  snow.season.model.calib.rel[is.na(snow.season.model.calib.rel)]<-0
  snow.season.corrected.calib.rel[is.na(snow.season.corrected.calib.rel)]<-0
# 
#   vioplot::vioplot(snow.season.model.calib.rel, at=seq(from=1,length.out=data.length, by=4),
#                    col=5, colMed =1, ylab="Rel. error (-)", main = "Calibration, relative error",yalim=c(-1,1),
#                    names = calib.years, ylim=c(-1,1))
#   vioplot::vioplot(snow.season.corrected.calib.rel, at = seq(from = 2, length.out = data.length, by = 4),
#                    col = 6, add = TRUE, colMed =1)
#   legend("topleft", c("model","corrected"), col=c(5,6), lwd=7,
#          bty="n", ncol=3)
#   legend("bottomleft", c(paste("model mean",
#                                round(mean(snow.season.model.calib.rel),2)),
#                          paste("model mae",
#                                round(mean(abs(snow.season.model.calib.rel)),2)),
#                          paste("corrected mean",
#                                round(mean(snow.season.corrected.calib.rel),2)),
#                          paste("corrected mae",
#                                round(mean(abs(snow.season.corrected.calib.rel)),2))),
#          bty="n", ncol=3)

  #Plot absolute values validation
  valid.years <- colnames(snow.season.training.valid)
  data.length <- length(valid.years)
  # vioplot::vioplot(snow.season.training.valid, at=seq(from=1,length.out=data.length, by=4),
  #                  col=2, colMed =1, ylab="Snow season length (day)", main = "Validation",
  #                  names = valid.years)
  # legend("topleft", c("training","model","corrected"), col=c(2,5,6), lwd=7,
  #        bty="n", ncol = 1)
  # vioplot::vioplot(snow.season.model.valid, at = seq(from = 2,length.out = data.length, by = 4),
  #                  col = 5, add = TRUE, colMed =1)
  # vioplot::vioplot(snow.season.corrected.valid, at = seq(from = 3, length.out = data.length, by = 4),
  #                  col = 6, add = TRUE, colMed =1)

  snow.season.model.valid.rel <- (snow.season.model.valid-snow.season.training.valid)/snow.season.training.valid
  snow.season.corrected.valid.rel <- (snow.season.corrected.valid-snow.season.training.valid)/snow.season.training.valid
  snow.season.model.valid.rel[snow.season.model.valid.rel>1]<-1
  snow.season.corrected.valid.rel[snow.season.corrected.valid.rel>1]<-1
  snow.season.model.valid.rel[is.na(snow.season.model.valid.rel)]<-0
  snow.season.corrected.valid.rel[is.na(snow.season.corrected.valid.rel)]<-0

  #Plot relative error validation
  # vioplot::vioplot(snow.season.model.valid.rel, at=seq(from=1,length.out=data.length, by=4),
  #                  col=5, colMed =1, ylab="Snow season length (day)", main = "Validation",yalim=c(-1,1),
  #                  names = valid.years, ylim=c(-1,1))
  # vioplot::vioplot(snow.season.corrected.valid.rel, at = seq(from = 2, length.out = data.length, by = 4),
  #                  col = 6, add = TRUE, colMed =1)
  # legend("topleft", c("model","corrected"), col=c(5,6), lwd=7,
  #        bty="n", ncol=3)
  # legend("bottomleft", c(paste("model mean",
  #                              round(mean(snow.season.model.valid.rel),2)),
  #                        paste("model mae",
  #                              round(mean(abs(snow.season.model.valid.rel)),2)),
  #                        paste("corrected mean",
  #                              round(mean(snow.season.corrected.valid.rel),2)),
  #                        paste("corrected mae",
  #                              round(mean(abs(snow.season.corrected.valid.rel)),2))),
  #        bty="n", ncol=3)

  #Prepare output data
  out=list()
  out$model$calib$mean.error <- mean(snow.season.model.calib.rel)
  out$model$calib$mae <- mean(abs(snow.season.model.calib.rel))
  out$corrected$calib$mean.error <- mean(snow.season.corrected.calib.rel)
  out$corrected$calib$mae <- mean(abs(snow.season.corrected.calib.rel))
  out$model$valid$mean.error <- mean(snow.season.model.valid.rel)
  out$model$valid$mae <- mean(abs(snow.season.model.valid.rel))
  out$corrected$valid$mean.error <- mean(snow.season.corrected.valid.rel)
  out$corrected$valid$mae <- mean(abs(snow.season.corrected.valid.rel))

  #Add data to produce map plots from output
  data.training.calib$snow.season <- matrix(rowMeans(snow.season.training.calib))
  data.training.calib$data <- NULL
  out$data.training.calib <- data.training.calib
  data.model.calib$snow.season <- matrix(rowMeans(snow.season.model.calib))
  data.model.calib$data <- NULL
  out$data.model.calib <- data.model.calib
  data.corrected.calib$snow.season <- matrix(rowMeans(snow.season.corrected.calib))
  data.corrected.calib$data <- NULL
  out$data.corrected.calib <- data.corrected.calib

  data.training.valid$snow.season <- matrix(rowMeans(snow.season.training.valid))
  data.training.valid$data <- NULL
  out$data.training.valid <- data.training.valid
  data.model.valid$snow.season <- matrix(rowMeans(snow.season.model.valid))
  data.model.valid$data <- NULL
  out$data.model.valid <- data.model.valid
  data.corrected.valid$snow.season <- matrix(rowMeans(snow.season.corrected.valid))
  data.corrected.valid$data <- NULL
  out$data.corrected.valid <- data.corrected.valid

  return(out)
}


plt.snow.season.maps.calib <- function(snow.season.training.calib, snow.season.model.calib,
                                       snow.season.corrected.calib, snow.season.training.valid,
                                       snow.season.model.valid, snow.season.corrected.valid,
                                       out.plot.dir) {

  pdf(paste0(out.plot.dir,"/snow_season_maps.pdf"),width=12,height=8)

  par(mar=c(3,4,2,7), mgp=c(1.7,0.5,0),mfrow=c(2,3))

  #Plot map of absoulte values
  plot(snow.season.training.calib, var="snow.season", zlim=c(0,366),
       main="training calib", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.model.calib, var="snow.season", zlim=c(0,366),
       main="model calib", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.corrected.calib, var="snow.season", zlim=c(0,366),
       main="Corrected calib", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.training.valid, var="snow.season", zlim=c(0,366),
       main="training valid", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.model.valid, var="snow.season", zlim=c(0,366),
       main="model valid", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)
  plot(snow.season.corrected.valid, var="snow.season", zlim=c(0,366),
       main="Corrected valid", legend.lab="Snow season length (days)",
       legend.cex=0.7, legend.line=2.7)


  par(mar=c(3,4,2,8), mgp=c(1.7,0.5,0),mfrow=c(2,2))
  #Compute relative error
  snow.season.model.calib$snow.season.rel <- (snow.season.model.calib$snow.season -
                                                   snow.season.training.calib$snow.season) /
                                                  snow.season.training.calib$snow.season
  snow.season.corrected.calib$snow.season.rel <- (snow.season.corrected.calib$snow.season -
                                                    snow.season.training.calib$snow.season) /
                                                    snow.season.training.calib$snow.season
  snow.season.model.calib$snow.season.rel[snow.season.model.calib$snow.season.rel>1] <- 1
  snow.season.corrected.calib$snow.season.rel[snow.season.corrected.calib$snow.season.rel>1] <- 1

  snow.season.model.valid$snow.season.rel <- (snow.season.model.valid$snow.season -
                                                   snow.season.training.valid$snow.season) /
                                                   snow.season.training.valid$snow.season
  snow.season.corrected.valid$snow.season.rel <- (snow.season.corrected.valid$snow.season -
                                                    snow.season.training.valid$snow.season) /
                                                    snow.season.training.valid$snow.season
  snow.season.model.valid$snow.season.rel[snow.season.model.valid$snow.season.rel>1]<-1
  snow.season.corrected.valid$snow.season.rel[snow.season.corrected.valid$snow.season.rel>1]<-1

  #Plot relative error
  plot(snow.season.model.calib,var="snow.season.rel", zlim=c(-1,1),
       main="model calib", legend.lab="Error (%)",
       legend.cex=0.9, legend.line=2.7)
  plot(snow.season.corrected.calib,var="snow.season.rel", zlim=c(-1,1),
       main="Corrected calib", legend.lab="Error (%)",
       legend.cex=0.9, legend.line=2.7)
  plot(snow.season.model.valid,var="snow.season.rel", zlim=c(-1,1),
       main="model valid", legend.lab="Error (%)",
       legend.cex=0.9, legend.line=2.7)
  plot(snow.season.corrected.valid,var="snow.season.rel", zlim=c(-1,1),
       main="Corrected valid", legend.lab="Error (%)",
       legend.cex=0.9, legend.line=2.7)
  dev.off()

}
