################################################################################
############################# INNER FUNCTIONS ##################################
################################################################################

get.correlation <- function(data,indices, ncores) {
  data.tmp <- rowMeanCpp(data$data, ncores)
  diff=rep(0,times=length(data$x.indices))
  for (pixel in  1:length(data$x.indices)) {
    diff[pixel] = data.tmp[pixel]/mean(data.tmp[indices[[pixel]]])
  }
  return(matrix(diff))
}

################################################################################
############################# SINGLE MODEL RUN #################################
################################################################################
#' XXXXX
#'
#' \code{plt.snow.days.comparison} XXX
#' \code{\link{barplot()}}) XXXX
#'
#' @param data.training The training dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.model The model dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.corrected the corrected dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param out.plot.dir Path in which the RDS files should be saved. Three files are
#' created: snow_season_training, snow_season_model, snow_season_corrected
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param swe.threshold The threshold SWE value to define snow days
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
plt.spatial.correlation <- function(data.training, data.model,
                                    data.corrected, out.plot.dir,
                                    ncores = 1, dist.threshold = 5) {

  x.coords.vect <- data.training$x.coords[data.training$x.indices]/1000
  y.coords.vect <- data.training$y.coords[data.training$y.indices]/1000

  indices = spatialCorrelationCpp(x.coords.vect, y.coords.vect, ncores, dist.threshold)

  data.training$correlation=get.correlation(data.training, indices, ncores)
  data.model$correlation=get.correlation(data.model, indices, ncores)
  data.corrected$correlation=get.correlation(data.corrected, indices, ncores)

  pdf(paste0(out.plot.dir,"/snow_height_spatial_correlation.pdf"),width=12,height=8)
  par(mar=c(3,3,2,5), mgp=c(1.7,0.5,0),mfrow=c(2,3))

  plot(data.training,var="correlation", main=paste("training data"),zlim=c(0,2))
  plot(data.model,var="correlation", main=paste("Traning data"),zlim=c(0,2))
  plot(data.corrected,var="correlation", main=paste("Corrected data"),zlim=c(0,2))

  data.model$correlation.rel <- data.model$correlation-data.training$correlation
  data.corrected$correlation.rel <- data.corrected$correlation-data.training$correlation

  par(mfrow=c(1,2))
  out = list()
  out$model$calib$me <- round(mean(data.model$correlation.rel),4)
  out$model$calib$mae <- round(mean(abs(data.model$correlation.rel)),3)
  plot(data.model,var="correlation.rel",
       main=paste("Traning data", out$model$calib$me, out$model$calib$mae),
       zlim=c(-1,1))
  out$corrected$calib$me <- round(mean(data.corrected$correlation.rel),4)
  out$corrected$calib$mae <- round(mean(abs(data.corrected$correlation.rel)),3)
  plot(data.corrected,var="correlation.rel",
       main=paste("Corrected data", out$corrected$calib$me, out$corrected$calib$mae),
       zlim=c(-1,1))

  dev.off()
  return(out)
}
################################################################################
####################### CALIBRATION AND VALIDATION #############################
################################################################################
#' XXXXX
#'
#' \code{plt.snow.days.comparison} XXX
#' \code{\link{barplot()}}) XXXX
#'
#' @param data.training The training dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.model The model dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param data.corrected the corrected dataset (a vectdata object, see
#' \code{\link{vectorize.list()}})
#' @param out.plot.dir Path in which the RDS files should be saved. Three files are
#' created: snow_season_training, snow_season_model, snow_season_corrected
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#' @param swe.threshold The threshold SWE value to define snow days
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
plt.spatial.correlation.calib <- function(data.training.calib, data.model.calib,
                                          data.corrected.calib, data.training.valid,
                                          data.model.valid, data.corrected.valid,
                                          out.plot.dir, ncores = 1, dist.threshold = 5) {

  x.coords.vect <- data.training.calib$x.coords[data.training.calib$x.indices]/1000
  y.coords.vect <- data.training.calib$y.coords[data.training.calib$y.indices]/1000

  indices = spatialCorrelationCpp(x.coords.vect, y.coords.vect, ncores, dist.threshold)

  data.training.calib$correlation=get.correlation(data.training.calib, indices, ncores)
  data.training.valid$correlation=get.correlation(data.training.valid, indices, ncores)
  data.model.calib$correlation=get.correlation(data.model.calib, indices, ncores)
  data.model.valid$correlation=get.correlation(data.model.valid, indices, ncores)
  data.corrected.calib$correlation=get.correlation(data.corrected.calib, indices, ncores)
  data.corrected.valid$correlation=get.correlation(data.corrected.valid, indices, ncores)

  pdf(paste0(out.plot.dir,"/snow_height_spatial_correlation.pdf"),width=12,height=8)
  par(mar=c(3,3,2,5), mgp=c(1.7,0.5,0),mfrow=c(2,3))

  plot(data.training.calib,var="correlation", main=paste("training data calib"),zlim=c(0,2))
  plot(data.model.calib,var="correlation", main=paste("Traning data calib"),zlim=c(0,2))
  plot(data.corrected.calib,var="correlation", main=paste("Corrected data calib"),zlim=c(0,2))
  plot(data.training.valid,var="correlation", main=paste("training data valid"),zlim=c(0,2))
  plot(data.model.valid,var="correlation", main=paste("model data valid"),zlim=c(0,2))
  plot(data.corrected.valid,var="correlation", main=paste("Corrected data valid"),zlim=c(0,2))

  data.model.calib$correlation.rel <- data.model.calib$correlation-data.training.calib$correlation
  data.corrected.calib$correlation.rel <- data.corrected.calib$correlation-data.training.calib$correlation
  data.model.valid$correlation.rel <- data.model.valid$correlation-data.training.valid$correlation
  data.corrected.valid$correlation.rel <- data.corrected.valid$correlation-data.training.valid$correlation

  out = list()
  par(mfrow=c(2,2))

  out$model$calib$me <- round(mean(data.model.calib$correlation.rel),4)
  out$model$calib$mae <- round(mean(abs(data.model.calib$correlation.rel)),3)
  plot(data.model.calib,var="correlation.rel",
       main=paste("Traning data calib", out$model$calib$me, out$model$calib$mae),
       zlim=c(-1,1))
  out$corrected$calib$me <- round(mean(data.corrected.calib$correlation.rel),4)
  out$corrected$calib$mae <- round(mean(abs(data.corrected.calib$correlation.rel)),3)
  plot(data.corrected.calib,var="correlation.rel",
       main=paste("Corrected data calib", out$corrected$calib$me, out$corrected$calib$mae),
       zlim=c(-1,1))
  out$model$valid$me <- round(mean(data.model.valid$correlation.rel),4)
  out$model$valid$mae <- round(mean(abs(data.model.valid$correlation.rel)),3)
  plot(data.model.valid,var="correlation.rel",
       main=paste("Traning data valid", out$model$valid$me, out$model$valid$mae),
       zlim=c(-1,1))
  out$corrected$valid$me <- round(mean(data.corrected.valid$correlation.rel),4)
  out$corrected$valid$mae <- round(mean(abs(data.corrected.valid$correlation.rel)),3)
  plot(data.corrected.valid,var="correlation.rel",
       main=paste("Corrected data valid", out$corrected$valid$me, out$corrected$valid$mae),
       zlim=c(-1,1))

  dev.off()

  return(out)
}