#' read.swe
#' 
#' `read.swe` reads SWE data from a NetCDF file and returns a vectdata
#' object containing all the data (see [vectorize.list()] for details 
#' about these objects).
#'
#' @param data.path path of the NetCDF file
#' @param start timestep at which the reading should start 
#' @param end timestep at which the reading should end (mutually exclusive with count)
#' @param count number of timestep to read (mutually exclusive with end)
#' @param data.dim name of the dimension containing the SWE data in the NetCDF file
#' @param x.coords.dim name of the dimension containing the x coordinates in the 
#' NetCDF file
#' @param y.coords.dim name of the dimension containing the y coordinates in the
#' NetCDF file
#' @param mask mask (in vectdatamask format, see [get.mask.na.vectdata()]
#' for details about these objects) to apply to the data
#' @param dem.path path of the DEM file (in ASCII grid) to use, should use the
#' same coordinate system as the NetCDF file
#' @param x.shift shift in x direction to apply to the DEM (in x.coords.dim unit)
#' @param y.shift shift in y direction to apply to the DEM (in y.coords.dim unit)
#' @param curvature.scale distance to be used to compute the curvature (in the
#' same units as in the NetCDF file)
#' @param silent should debug output be printed 
#' 
#' @return a vectdata object containing all the data (see [vectorize.list()] for 
#' details about these objects).
#'
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)
#' @export
read.swe <- function (data.path, start=NULL, end=NULL, count=NULL,
        data.dim="SWE", x.coords.dim="chx", y.coords.dim="chy", mask=NULL,
        dem.path, x.shift=0, y.shift=0, curvature.scale = NA,
        silent=FALSE) {

  ncdata <- read.nc.data(data.path, start, end, count, silent, data.dim,
                         x.coords.dim, y.coords.dim)
  vectdata <- vectorize.list(ncdata)

  # Force memory cleaning
  rm(ncdata)
  gc()

  gis.data <- get.dem.slope.aspect.curvature(dem.path, x_shift = 0, y_shift = 0, curvature.scale)
  dem <- vectorize.list(gis.data$dem)
  slope <- vectorize.list(gis.data$slope)
  aspect <- vectorize.list(gis.data$aspect)
  curvature <- vectorize.list(gis.data$curvature)

  # Force memory cleaning
  rm(gis.data)
  gc()

  if(!identical(dem$x.coords,vectdata$x.coords) |
     !identical(dem$y.coords,vectdata$y.coords)){
    print("[W] Area and/or resolution of the dem and NetCCDF file differ")
  }

  coords.1 <- paste(vectdata$x.coords[vectdata$x.indices],
                    vectdata$y.coords[vectdata$y.indices], sep="_")
  coords.2 <- paste(dem$x.coords[dem$x.indices], dem$y.coords[dem$y.indices],
                    sep="_")
  vectdata$dem <- as.matrix(dem$data[match(coords.1, coords.2)])
  coords.2 <- paste(aspect$x.coords[aspect$x.indices], aspect$y.coords[aspect$y.indices],
                    sep="_")
  vectdata$aspect <- as.matrix(aspect$data[match(coords.1, coords.2)])
  coords.2 <- paste(slope$x.coords[slope$x.indices], slope$y.coords[slope$y.indices],
                    sep="_")
  vectdata$slope <- as.matrix(slope$data[match(coords.1, coords.2)])
  coords.2 <- paste(slope$x.coords[slope$x.indices], slope$y.coords[slope$y.indices],
                    sep="_")
  vectdata$curvature <- as.matrix(curvature$data[match(coords.1, coords.2)])

  if(!is.null(mask)){
    print("[i] Start masking the data")
    vectdata <- mask.vectdata(vectdata, mask)
    print("[i] Finished masking the data")
  }
  return(vectdata)
}


#' read.nc.data
#' 
#' `read.nc.data` reads SWE data from a NetCDF file and returns a list 
#' containing teh data
#'
#' @param file.path path of the NetCDF file
#' @param start timestep at which the reading should start 
#' @param end timestep at which the reading should end (mutually exclusive with count)
#' @param count number of timestep to read (mutually exclusive with end)
#' @param silent should debug output be printed 
#' @param data.dim name of the dimension containing the SWE data in the NetCDF file
#' @param x.coords.dim name of the dimension containing the x coordinates in the 
#' NetCDF file
#' @param y.coords.dim name of the dimension containing the y coordinates in the
#' NetCDF file
#' 
#' @return A list with entries `x.coords`, `y.coords`, `time` (in the  same format
#' as in te NetCDF file), `time_date` (in [Dates] format), and `data` (containing
#' the SWE data as a 3D matrix)
#'
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)
#' @noRd
read.nc.data <- function(file.path, start=NULL, end=NULL, count=NULL,
                         silent=FALSE, data.dim="SWE", x.coords.dim="chx",
                         y.coords.dim="chy"){
  if(!is.null(end) && !is.null(count)) {
    stop("Only one of 'end' or 'count' should be provided")
  }
  if(!silent){print(paste0("[i] Start reading: ",file.path))}
  nc <- ncdf4::nc_open(file.path)
  ncdata <- list()
  ncdata$x.coords <- ncdf4::ncvar_get(nc, x.coords.dim)
  ncdata$y.coords <- ncdf4::ncvar_get(nc, y.coords.dim)
  ncdata$time <- ncdf4::ncvar_get(nc, "time")
  ncdata$time.date <- round(as.POSIXct(ncdf4.helpers::nc.get.time.series(nc,
                                              time.dim.name="time")),"days")

  if ((!is.null(start) || !is.null(end)) && is.null(count)) {
    subset <- cut.time.start.end(ncdata$time.date, start,end)
    ncdata$time <- ncdata$time[subset]
    ncdata$time.date <- ncdata$time.date[subset]
    ncdata$data <- ncdf4::ncvar_get(nc,data.dim, start=c(1,1,subset[1]),
                                    count=c(length(ncdata$x.coords),
                                            length(ncdata$y.coords),
                                            length(subset)))
  } else if (!is.null(count)) {
    subset <- cut.time.start.count(ncdata$time.date, start, count)
    ncdata$time <- ncdata$time[subset]
    ncdata$time.date <- ncdata$time.date[subset]
    ncdata$data <- ncdf4::ncvar_get(nc, data.dim, start=c(1,1,subset[1]),
                                           count=c(length(ncdata$x.coords),
                                                   length(ncdata$y.coords),
                                                   length(subset)))
  } else {
    ncdata$data <- ncdf4::ncvar_get(nc, data.dim)
  }
  ncdf4::nc_close(nc)
  if(!silent){print(paste0("[i] Finished reading: ",file.path))}
  return(ncdata)
}



cut.time.start.end<-function(timestamps, start=NULL, end=NULL)
{
  if(is.null(start) && is.null(end)) {
    stop("At least one of 'start' or 'end' argument must be provided")
  }
  if(is.null(start)){start.ts <- timestamps[1]}
  else if(is.numeric(start)){start=timestamps[start]}
  else{start = round(as.POSIXct(start,tz="GMT"),"days")}
  if(is.null(end)){end <- timestamps[length(timestamps)]}
  else if(is.numeric(end)){start <- timestamps[end]}
  else{end = round(as.POSIXct(end,tz="GMT"),"days")}
  if(start>=end) {
    stop("'start' should be < 'end'")
  }
  subset = which(timestamps>=start & timestamps<=end)
  return(subset)
}

cut.time.start.count <- function(timestamps, start=NULL, count=NULL)
{
  if(is.null(start) && is.null(count)) {
    stop("At least one of 'start' or 'end' argument must be provided")
  }
  if(is.null(start)){start.ts <- timestamps[1]}
  else if(is.numeric(start)){start <- timestamps[start]}
  else{start = round(as.POSIXct(start, tz="GMT"), "days")}
  subset = which(timestamps >= start)
  if(length(subset) < count) {
    stop("Start + count > data length")
  }
  subset <- subset[1:count]
  return(subset)
}

#' read.asc
#' 
#' `read.asc` reads a grid in ASCII format and returns it as a [list].
#'
#' @param filename path of the ASCII file
#' @param x.shift shift in x direction to apply to the data (in x.coords.dim unit)
#' @param y.shift shift in y direction to apply to the data (in y.coords.dim unit)
#' 
#' @return A list with entries `x.coords`, `y.coords` and `data` (containing
#' the data as a 2D matrix)
#'
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)
#' @export
read.asc <- function(filename, x.shift=0, y.shift=0, ...)
{
  print(paste("[I] Reading ASC file:", filename))
  header.file <- file(filename, "r")
  lines <- readLines(header.file, n = 10)
  close(header.file)
  header <- list()
  n <- 0
  for (line in lines) {
    line=tolower(line)
    if (startsWith(line,"ncols") || startsWith(line,"nrows") ||
        startsWith(line,"xllcorner") || startsWith(line,"yllcorner") ||
        startsWith(line,"cellsize") || startsWith(line,"nodata_value")) {
      header.element <- strsplit(line,c("\\s+","\t"))[[1]]
      header[[header.element[1]]] <- as.numeric(header.element[2])
    } else {
      break
    }
    n <- n+1
  }

  data.list <- list()
  data.list$data <- t(apply(as.matrix(read.table(filename, skip = n)), 2, rev))
  na.value <- as.numeric(ifelse("nodata_value" %in% names(header), header$nodata_value, NA))
  if(!is.na(na.value)){
    data.list$data[data.list$data == na.value] <- NA
  }
  data.list$x.coords <- header$xllcorner + x.shift +
                      seq.int(from = 0, length.out = header$ncols)*header$cellsize
  data.list$y.coords <- header$yllcorner + y.shift +
    seq.int(from = 0, length.out = header$nrows)*header$cellsize

  return(data.list)
}


#' read.tif
#' 
#' `read.tif` reads a grid in GetoTIFF format and returns it as a [list].
#'
#' @param filename path of the GetoTIFF file
#' @param x.shift shift in x direction to apply to the data (in x.coords.dim unit)
#' @param y.shift shift in y direction to apply to the data (in y.coords.dim unit)
#' 
#' @return A list with entries `x.coords`, `y.coords` and `data` (containing
#' the data as a 2D matrix)
#'
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)
#' @export
read.tif <- function(filename, x.shift=0, y.shift=0, ...)
{
  print(paste("[I] Reading GeoTIFF file:",filename))
  data=raster::raster(filename)
  if(raster::res(data)[1] != raster::res(data)[2]) {
    stop("Only squared pixels are supported for GeoTIFF file")
  }
  data.list <- list()
  data.list$data <- t(apply(raster::as.matrix(data),2,rev))
  data.list$x.coords <- raster::extent(data)[1] + x.shift +
    seq.int(from = 0, length.out = dim(data)[2])*raster::res(data)[1]
  data.list$y.coords <- raster::extent(data)[3] + y.shift +
    seq.int(from = 0, length.out = dim(data)[1])*raster::res(data)[2]
  return(data.list)
}


#' save.netcdf
#' 
#' `save.netcdf` convert a vectdata object (see [vectorize.list()] for details 
#' about these objects) into a NetCDF file. Only coordiantes and SWE are 
#' written
#'
#' @param vectdata vectdata object
#' @param path pate of the output NetCDF file
#'
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)
#' @export
save.netcdf <- function(vectdata, path) {

  chx <- ncdf4::ncdim_def(name="chx", longname="easting", units="m",
                          vals=vectdata$x.coords)

  chy <- ncdf4::ncdim_def(name="chy", longname="northing", units="m",
                          vals=vectdata$y.coords)

  time <- ncdf4::ncdim_def(name="time",units="hours since 1960-01-01",
                           vals=vectdata$time , unlim=TRUE)

  mv <- -999 #missing value to use
  var_SWE <- ncdf4::ncvar_def("SWE", "m", list(chx, chy, time),
                              longname="SWE from OSHD corrected with SnowQM", mv)

  nc <- ncdf4::nc_create(path, list(var_SWE))

  ncdf4::ncatt_put(nc, 0, "Title", "SWE from OSHD corrected with SnowQM")
  ncdf4::ncatt_put(nc, 0, "Source", "Created by Adrien Michel, WSL/SLF and MeteoSwiss, 2022")
  ncdf4::ncatt_put(nc, 0, "References", "SnowQM 1.0: A fast R Package for snow water equivalent maps bias-correction using quantile mapping")
  ncdf4::ncatt_put(nc, 0, "Created on", date())


  for (t in 1:length(vectdata$time)) {
    print(t)
    data.write <- matrix(NA, nrow=length(vectdata$x.coords),
                        ncol=length(vectdata$y.coords))

    for (i in seq_along(vectdata$x.indices)) {
      data.write[vectdata$x.indices[i], vectdata$y.indices[i]] <-
        vectdata$data[i,t]
    }
    ncdf4::ncvar_put(nc = nc,
                     varid = var_SWE,
                     vals = data.write,
                     start = c(1, 1, t),
                     count = c(-1, -1, 1))
  }
}





