#' get.overlap.netcdf
#' 
#' Computes the overlap between two NetCDF datasets.
#'
#' @param file1 path of the first NetCDF file
#' @param file1 Path of the second NetCDF file
#' @param data.dim variable to use to compute the overlap
#' @param timestep timestep to use to compute the overlap
#' @param x.coords.dim name of the dimension containing the x coordinates in the 
#' NetCDF file
#' @param y.coords.dim name of the dimension containing the y coordinates in the
#' NetCDF file
#' 
#' @return mask (in vectdatamask format, see [get.mask.na.vectdata()]
#' for details about these objects) of teh overlapping pixels
#'  
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)
#' @export
get.overlap.netcdf <- function(file1, file2, data.dim="SWE", timestep,
                                x.coords.dim="chx", y.coords.dim="chy")
{
  data1 <- read.nc.data(file1, start=timestep, count=1, data.dim=data.dim,
                        x.coords.dim=x.coords.dim, y.coords.dim=y.coords.dim)
  data2 <- read.nc.data(file2, start=timestep, count=1, data.dim=data.dim,
                        x.coords.dim=x.coords.dim, y.coords.dim=y.coords.dim)

  vectdata1 <- vectorize.list(data1,"data")
  vectdata2 <- vectorize.list(data2,"data")

  mask1 <- get.mask.na(vectdata1)
  mask2 <- get.mask.na(vectdata2)

  mask <- get.ovelap.mask(mask1,mask2)

  return(mask)
}

#' get.mask.na
#'
#' Overload for [get.mask.na.vectdata()]
#' 
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)
#' @export
get.mask.na <- function (x, ...) {
  UseMethod("get.mask.na")
}

#' get.mask.na.vectdata
#'
#' Compute a mask (in vectdatamask format, see [get.mask.na.vectdata()]
#' for details about these objects) based on NA in a given vectdata object 
#' (see [vectorize.list()] for details about these objects).
#'
#' @param vectdata a vectdata object
#' 
#' @return a vectdatamask object
#' 
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)
#' @export
get.mask.na.vectdata<-function(vectdata)
{
  # Copy indices and coordinates from input
  mask <- structure(list(), class = c("vectdatamask","list"))
  mask[c("x.coords","y.coords","x.indices","y.indices")] <-
    vectdata[c("x.coords","y.coords","x.indices","y.indices")]

  indices.to.kepp=which(!is.na(vectdata$data))

  # Reduce vectors
  mask$x.indices <- mask$x.indices[indices.to.kepp]
  mask$y.indices <- mask$y.indices[indices.to.kepp]

  return(mask)
}

#' get.ovelap.mask
#' 
#' Compute a mask (in vectdatamask format, see [get.mask.na.vectdata()]
#' for details about these objects) based on the overlap between any
#' number of mask given as input.
#'
#' @param mask1 a vectdatamask object
#' @param mask2 a vectdatamask object
#' @param ... any number of additional vectdatamask objects
#' 
#' @return a vectdatamask object
#' 
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)
#' @export
get.ovelap.mask <- function(mask1, mask2, ...) {
  arguments <- list(...)

  if(!("vectdatamask" %in% class(mask1))){
    stop("All inputs should be of class vectdatamask")
  }
  if(!("vectdatamask" %in% class(mask2))){
    stop("All inputs should be of class vectdatamask")
  }

  # Concatenate x and y coords in a single vector
  coords.1 <- paste(mask1$x.coords[mask1$x.indices], mask1$y.coords[mask1$y.indices],
                    sep="_")
  coords.2 <- paste(mask2$x.coords[mask2$x.indices], mask2$y.coords[mask2$y.indices],
                    sep="_")

  # Check for match and keep only match
  indices.to.kepp <- sort(match(coords.1,coords.2))
  mask <- mask2
  mask$x.indices <- mask$x.indices[indices.to.kepp]
  mask$y.indices <- mask$y.indices[indices.to.kepp]

  if(length(arguments)>0)
  {
    mask=get.ovelap.mask(mask,...)
  }
  return(mask)
}

#' get.filter.mask.vectdata
#'
#' Compute a mask (in vectdatamask format, see [get.mask.na.vectdata()]
#' for details about these objects) based on elevation and snow towers
#' in a given vectdata object (see [vectorize.list()] for details about these 
#' objects).
#' 
#' Snow towers are defined by pixel having on average a SWE value greater than
#' a given threshold on teh 31st of August.
#' 
#' @param vectdata a vectdata object
#' @param swe.threshold threshold (in mSWE) to define snow towers
#' @param elevation.limit elevation limit above which pixels are masked
#' 
#' @return a vectdatamask object
#' 
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)
#' @export
get.filter.mask.vectdata <- function(vectdata, swe.threshold=NULL, elevation.limit=NULL)
{
  mask <- structure(list(), class = c("vectdatamask","list"))
  mask[c("x.coords","y.coords","x.indices","y.indices")] <-
    vectdata[c("x.coords","y.coords","x.indices","y.indices")]
  to.remove <- c()
  if(!is.null(swe.threshold)){
    times <- which(lubridate::month(vectdata$time.date)==8 & lubridate::day(vectdata$time.date)==31)
    snow.towers <- rowSums(vectdata$data[,times],na.rm = T)/22
    snow.towers <- ifelse(snow.towers < swe.threshold, FALSE, TRUE)
    to.remove <- c(to.remove, which(snow.towers))
  }
  if(!is.null(elevation.limit)){
    to.remove <- c(to.remove,which(vectdata$dem>elevation.limit))
  }
  to.remove <- unique(to.remove)
  if(length(to.remove) > 0){
    print(paste0("  [i] Removing: ",
                 round(length(to.remove)/length(mask$x.indices)*100,1),"% of pixels"))
    mask$x.indices <- mask$x.indices[-to.remove]
    mask$y.indices <- mask$y.indices[-to.remove]
  }
  return(mask)
}

set.na.vectdata <- function(vectdata, val.to.remove){
  vectdata$data[vectdata$data == val.to.remove] <- NA
  return(vectdata)
}


#' read.mask.asc
#' 
#' Compute a mask (in vectdatamask format, see [get.mask.na.vectdata()]
#' for details about these objects) based on the values in an ASCII file. All
#' NA values and all values equal to the parameter `val.to.remove` are used to
#' define the masked area
#'
#' @param filename path of the ASCII file
#' @param val.to.remove values to be used in adition to NA to define teh mask
#' @param x.shift shift in x direction to apply to the data (in x.coords.dim unit)
#' @param y.shift shift in y direction to apply to the data (in y.coords.dim unit)
#' 
#' @return a vectdatamask object
#'
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)
#' @export
read.mask.asc <- function(filename, val.to.remove=NULL, x.shift=0,
                          y.shift=0){
  asc.list <- read.asc(filename, x.shift, y.shift)
  mask.vectdata <- vectorize.list(asc.list, "data")
  if(!is.null(val.to.remove)){
    mask.vectdata <- set.na.vectdata(mask.vectdata, val.to.remove)
  }
  mask <- get.mask.na(mask.vectdata)
  return(mask)
}

#' read.mask.tif
#' 
#' Compute a mask (in vectdatamask format, see [get.mask.na.vectdata()]
#' for details about these objects) based on the values in an GeoTIFF file. All
#' NA values and all values equal to the parameter `val.to.remove` are used to
#' define the masked area
#'
#' @param filename path of the GeoTIFF file
#' @param val.to.remove values to be used in adition to NA to define teh mask
#' @param x.shift shift in x direction to apply to the data (in x.coords.dim unit)
#' @param y.shift shift in y direction to apply to the data (in y.coords.dim unit)
#' 
#' @return a vectdatamask object
#'
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)
#' @export
read.mask.tif <- function(filename, val.to.remove=NULL, x.shift=0,
                          y.shift=0){
  tif.list <- read.tif(filename, x.shift, y.shift)
  mask.vectdata <- vectorize.list(tif.list, "data")
  if(!is.null(val.to.remove)){
    mask.vectdata <- set.na.vectdata(mask.vectdata, val.to.remove)
  }
  mask <- get.mask.na(mask.vectdata)
  return(mask)
}

#' mask.vectdata
#' 
#' Mask a vectdata object (see [vectorize.list()] for details about these 
#' objects) based on a mask (in vectdatamask format, see [get.mask.na.vectdata()]
#' for details about these objects). All masked data are deleted.
#'
#' @param vectdata a vectdata object
#' @param mask a vectdatamask object
#' 
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)
#'
#' @export
mask.vectdata <- function(vectdata, mask)
{
  if(!("vectdatamask" %in% class(mask))) {
    stop("Sencond input should be of class vectdatamask")
  }

  coords.1 <- paste(vectdata$x.coords[vectdata$x.indices],
                    vectdata$y.coords[vectdata$y.indices], sep="_")
  coords.2 <- paste(mask$x.coords[mask$x.indices], mask$y.coords[mask$y.indices],
                    sep="_")
  to.keep <- sort(match(coords.2, coords.1))

  if(is.na(min(to.keep))) {
    print("[W] Some pixels in the mask are not found in the data to be masked")
  }

  print(paste0("[i] Masking ",
               round((length(vectdata$x.indices)-length(to.keep))/
                       length(vectdata$x.indices)*100,1),"% of the pixels"))

  if(length(to.keep) == 0) {
    error("Masking impossible, all the data would be removed")
  }

  for(var in names(vectdata)) {
    if(var %in% c("time","time.date","x.coords","y.coords")){
      next
    }
    if(length(dim(vectdata[[var]])) == 2) {
      vectdata[[var]] <- vectdata[[var]][to.keep,,drop=FALSE]
    } else{
      vectdata[[var]] <- vectdata[[var]][to.keep]
    }
  }
  return(vectdata)
}

