#' This file contains only internal functions to compute the slope, aspect,
#' and curvature from a DEM. This is a R implementation io the algorithms
#' of the meteoio library, see https://code.wsl.ch/snow-models/meteoio/-/blob/master/meteoio/dataClasses/DEMObject.cc
#' for details and references
#' 
#' @author Adrien Michel, 2022 (WSL/SLF and MeteoSwiss)


safe.get <- function(i, j, nx, ny) {
  i2 <- 1
  j2 <- 1
  if (i < 1) {
    i2 <- 1
  } else if (i > nx) {
    i2 <- nx
  } else {
    i2 <- i
  }
  if ( j < 1 ) {
    j2 <- 1
  } else if ( j > ny ) {
    j2 <- ny
  } else {
    j2 <- j;
  }
  return(c(i2, j2))
}


get.neighbours <- function(i, j, dem, nx, ny, cellsize, curvature.scale = NA) {

  if(is.na(curvature.scale)){
    step <- 1
  } else {
    max_step <- min(nx-1, ny-1);
    step <- floor(min(c(max_step, max(c(1, (curvature.scale / cellsize) + 0.5)))))
  }

  neighbours <- matrix(NA, ncol=3, nrow =3)
  if ( i > step && i <= (nx-step) && j > step && j <= (ny-step) ) {
    neighbours[1,1] <- dem[i-step, j+step]
    neighbours[1,2] <- dem[i, j+step]
    neighbours[1,3] <- dem[i+step, j+step]
    neighbours[2,1] <- dem[i-step, j]
    neighbours[2,2] <- dem[i, j]
    neighbours[2,3] <- dem[i+step, j]
    neighbours[3,1] <- dem[i-step, j-step]
    neighbours[3,2] <- dem[i, j-step]
    neighbours[3,3] <- dem[i+step, j-step]
  } else {
    indices <- safe.get(i-step, j+step, nx, ny)
    neighbours[1,1] <- dem[indices[1], indices[2]]
    indices <- safe.get(i, j+step, nx, ny)
    neighbours[1,2] <- dem[indices[1], indices[2]]
    indices <- safe.get(i+step, j+step, nx, ny)
    neighbours[1,3] <- dem[indices[1], indices[2]]
    indices <- safe.get(i-step, j, nx, ny)
    neighbours[2,1] <- dem[indices[1], indices[2]]
    indices <- safe.get(i, j, nx, ny)
    neighbours[2,2] <- dem[indices[1], indices[2]]
    indices <- safe.get(i+step, j, nx, ny)
    neighbours[2,3] <- dem[indices[1], indices[2]]
    indices <- safe.get(i-step, j-step, nx, ny)
    neighbours[3,1] <- dem[indices[1], indices[2]]
    indices <- safe.get(i, j-step, nx, ny)
    neighbours[3,2] <- dem[indices[1], indices[2]]
    indices <- safe.get(i+step, j-step, nx, ny)
    neighbours[3,3] <- dem[indices[1], indices[2]]
  }
  return(neighbours)
}


get.slope.corripio <- function(neighbours, cellsize) {
  out <- list()
  if ( !is.na(neighbours[1,1]) && !is.na(neighbours[1,3]) && !is.na(neighbours[3,1]) && !is.na(neighbours[3,3]) ) {
    vx <- (neighbours[3,1] + neighbours[1,1] - neighbours[3,3] - neighbours[1,3]) / (2.*2.*cellsize);
    vy <- (neighbours[3,1] - neighbours[1,1] + neighbours[3,3] - neighbours[1,3]) / (2.*2.*cellsize);
    out$slope <- atan( sqrt(vx*vx+vy*vy) ) * 180/pi
    out$vect <- c(vx, vy, 1)
  } else {
    out <- get.slope.hick(neighbours, cellsize)
  }
  return(out)
}

steepest.gradient <- function(neighbours, cellsize){
  smax <- -1.
  sqrt2 <- sqrt(2.)
  if (!is.na(neighbours[1,1]))
    smax <- max( smax, abs(neighbours[2,2] - neighbours[1,1])/(cellsize*sqrt2) )
  if (!is.na(neighbours[1,2]))
    smax <- max( smax, abs(neighbours[2,2] - neighbours[1,2])/(cellsize) )
  if (!is.na(neighbours[1,3]))
    smax <- max( smax, abs(neighbours[2,2] - neighbours[1,3])/(cellsize*sqrt2) )
  if (!is.na(neighbours[2,1]))
    smax <- max( smax, abs(neighbours[2,2] - neighbours[2,1])/(cellsize) )
  if (!is.na(neighbours[2,3]))
    smax <- max( smax, abs(neighbours[2,2] - neighbours[2,3])/(cellsize) )
  if (!is.na(neighbours[3,1]))
    smax <- max( smax, abs(neighbours[2,2] - neighbours[3,1])/(cellsize*sqrt2) )
  if (!is.na(neighbours[3,2]))
    smax <- max( smax, abs(neighbours[2,2] - neighbours[3,2])/(cellsize) )
  if (!is.na(neighbours[3,3]))
    smax <- max( smax, abs(neighbours[2,2] - neighbours[3,3])/(cellsize*sqrt2) )

  if (smax<0.) return(NA)

  return(smax)
}


line.gradient <- function(a1, a2, a3) {
  if (!is.na(a3) && !is.na(a1)) {
    return(a3 - a1)
  } else {
    if (!is.na(a2)) {
      if (!is.na(a3))
        return((a3 - a2)*2.)
      if (!is.na(a1))
        return((a2 - a1)*2.)
    }
  }
  return(NA)
}


fill.missing.gradient <- function(delta1, delta2) {
  if (!is.na(delta1) && !is.na(delta2)) {
    return(0.5*(delta1+delta2))
  } else {
    if (!is.na(delta1)) return(delta1)
    if (!is.na(delta2)) return(delta2)
  }
  return(NA)
}

surface.gradient <- function(neighbours) {

  dx1 <- line.gradient(neighbours[3,1], neighbours[3,2], neighbours[3,3])
  dx2 <- line.gradient(neighbours[2,1], neighbours[2,2], neighbours[2,3])
  dx3 <- line.gradient(neighbours[1,1], neighbours[1,2], neighbours[1,3])

  dy1 <- line.gradient(neighbours[3,1], neighbours[2,1], neighbours[1,1])
  dy2 <- line.gradient(neighbours[3,2], neighbours[2,2], neighbours[1,2])
  dy3 <- line.gradient(neighbours[3,3], neighbours[2,3], neighbours[1,3])

  if (is.na(dx1)) dx1 = fill.missing.gradient(dx2, dx3)
  if (is.na(dx2)) dx2 = fill.missing.gradient(dx1, dx3)
  if (is.na(dx3)) dx3 = fill.missing.gradient(dx1, dx2)
  if (is.na(dy1)) dy1 = fill.missing.gradient(dy2, dy3)
  if (is.na(dy2)) dy2 = fill.missing.gradient(dy1, dy3)
  if (is.na(dy3)) dy3 = fill.missing.gradient(dy1, dy2)

  if (!is.na(dx1) && !is.na(dy1)) {
    dx_sum <- (dx1 + 2.*dx2 + dx3) * 0.25
    dy_sum <- (dy1 + 2.*dy2 + dy3) * 0.25
  } else {
    dx_sum <- NA
    dy_sum <- NA
  }

  return(c(dx_sum, dy_sum))
}



get.slope.hick <- function(neighbours, cellsize) {
  smax = steepest.gradient(neighbours, cellsize)
  out = list()
  if (is.na(smax)) {
    out$slope <- NA;
    out$vect <- c(NA, NA,NA)
  } else {
    out$slope <- atan(smax) * 180/pi
    if ( smax > 0. ) {
      gradient <- surface.gradient(neighbours)
      if ( is.na(gradient[1]) || is.na(gradient[2]) ) {
        out$slope <- NA;
        out$vect <- c(NA, NA,NA)
      } else {
        vx <- -gradient[1] / (2. * cellsize)
        vy <- -gradient[2] / (2. * cellsize)
        out$vect <- c(vx, vy, 1)
      }
    } else {
      out$vect <- c(0, 0, 1)
    }
  }
  return(out)
}


get.aspect <- function(slope, vect) {

  if (is.na(vect[1]) || is.na(vect[2]) || is.na(vect[3]) || is.na(slope)) {
    return(NA)
  }
  if ( slope > 0. ) {
    if ( vect[1] == 0. ) {
      aspect <- if (vect[2] < 0.) 180. else 0.
      return(aspect)
    } else {
      angle_deg <- round( atan2(vect[2], vect[1]) * 180/pi * 10.) / 10
      aspect <- (90. - angle_deg + 360.)%%360.
      return(aspect)
    }
  } else {
    return(0)
  }
}

avg.height <- function(z1, z2, z3) {

  if (!is.na(z1) && !is.na(z3)) {
    return(0.5*(z1+z3))
  }
  if (!is.na(z1) && !is.na(z2)) {
    return(0.5*(z1+z2))
  }
  if (!is.na(z2) && !is.na(z3)) {
    return(0.5*(z3+z2))
  }

  return(NA)
}


get.curvature <- function(neighbours, cellsize, nx, ny, curvature.scale = NA) {
  if(is.na(curvature.scale)){
    step <- 1
  } else {
    max_step = min(nx-1, ny-1);
    step = floor(min(c(max_step, max(c(1, (curvature.scale / cellsize) + 0.5)))))
  }
  if (!is.na(neighbours[2,2])) {
    Zwe <- avg.height(neighbours[2,1], neighbours[2,2], neighbours[2,3])
    Zsn <- avg.height(neighbours[1,2], neighbours[2,2], neighbours[3,2])
    Zswne <- avg.height(neighbours[3,1], neighbours[2,2], neighbours[1,3])
    Znwse <- avg.height(neighbours[1,1], neighbours[2,2], neighbours[3,3])

    sqrt2 <- sqrt(2.)
    dX <- 2. * step * cellsize
    sum <- 0.
    count <- 0

    if (!is.na(Zwe)) {
      sum <- sum + (neighbours[2,2]-Zwe)/dX
      count <- count + 1
    }
    if (!is.na(Zsn)) {
      sum <- sum + (neighbours[2,2]-Zsn)/dX
      count <- count + 1
    }
    if (!is.na(Zswne)) {
      sum <- sum + (neighbours[2,2]-Zswne)/(dX*sqrt2)
      count <- count + 1
    }
    if (!is.na(Znwse)) {
      sum <- sum + (neighbours[2,2]-Znwse)/(dX*sqrt2)
      count <- count + 1
    }

    if (count != 0) return (1./count * sum)
  }
  return(NA)
}

get.dem.slope.aspect.curvature <- function(dem.path, x_shift = 0, y_shift = 0,
                                           curvature.scale = NA) {

  dem <- read.asc(dem.path, x_shift, y_shift)
  cellsize <- dem$x.coords[2]-dem$x.coords[1]
  nx <- dim(dem$data)[1]
  ny <- dim(dem$data)[2]
  if(is.na(curvature.scale)) {
    curvature.scale <- cellsize
  }
  slope = dem
  aspect = dem
  curvature = dem
  slope$data[,] <- NA
  aspect$data[,] <- NA
  curvature$data[,] <- NA
  for ( j in c(1:ny) ) {
    for ( i in c(1:nx) ) {
      if ( !is.na(dem$data[i,j]) ) {
        neighbours <- get.neighbours(i, j, dem$data, nx, ny, cellsize)
        slope.and.vect <- get.slope.corripio(neighbours, cellsize)
        slope$data[i,j] <- slope.and.vect$slope
        aspect$data[i,j] <- get.aspect(slope.and.vect$slope, slope.and.vect$vect)
        neighbours <- get.neighbours(i, j, dem$data, nx, ny, cellsize, curvature.scale)
        curvature$data[i,j] <- get.curvature(neighbours, nx, ny, cellsize, curvature.scale)
      }
    }
  }
  return(list(dem = dem, slope=slope,aspect=aspect,curvature=curvature))
}
