#' Apply the trained QM model to a given dataset
#'
#' \code{apply.model} applies a trained QM model (see
#' \code{\link{compute.quantiles()}}) to the dataset given as input.
#'
#' @param in.data.file Dataset on which the QM model should be applied. Dataset
#' should be a RDS file in a vectdata format (see \code{\link{vectorize.list()}})
#' @param out.data.file Path of the file to write the corrected dataset. Dataset
#' will be a RDS file in a vectdata format (see \code{\link{vectorize.list()}})
#' @param source.path Directory containing the directories with quantile distributions of the
#' training and target datasets (see \code{\link{vectorize.list()}})
#' @param ncores number of cores to be used
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
apply.model <- function(in.data.file, out.data.file, quantile.path, ncores, correct.first = TRUE)
{
  print("[I] Creating output directory")
  dir.create(dirname(out.data.file), showWarnings = FALSE)

  print("[I] Reading data")
  data <- readRDS(in.data.file)

  print("[I] Computing days of the year")
  doys=as.integer(strftime(data$time.date, format = "%j"))
  doys.unique=sort(unique(doys))
  `%do%` <- foreach::`%do%`
  doy.selection = foreach::foreach(doy=doys.unique) %do% {
    which(doys == doy)
  }
  names(doy.selection)=doys.unique

  print("[I] Correcting data")
  start.time <- Sys.time()
  corrected.swe <- applyQMCpp(data$data,
                              doyArray = doys,
                              xCoords=round(data$x.coords[data$x.indices]),
                              yCoords=round(data$y.coords[data$y.indices]),
                              modelPath = paste0(quantile.path,"/model_quantiles"),
                              trainingPath = paste0(quantile.path,"/training_quantiles"),
                              ncores = ncores,
                              correctFirst = correct.first)
  data.corrected <- data
  data.corrected$data <- matrix(corrected.swe, ncol=dim(data$data)[2])

  print(paste("[I] Correction computed in",
              as.numeric(round(difftime(Sys.time(),start.time, units = "mins"),2)),
              "minutes"))

  saveRDS(data.corrected,out.data.file)

  print("[I] Cleaning memory")
  rm(list=ls())
  invisible(gc())
}

