#' Produce plots to assess the downscaling quality
#'
#' \code{assessment.plot} produces two files (<out.plot.dir>/summary.pdf and
#' <out.plot.dir>/seasonnal_summary.pdf) to asses the quality of the downscaling.
#' The plots are produced using: \code{\link{plt.seasonnal.summary.details()}},
#' \code{\link{plt.seasonnal.summary()}}, \code{\link{plt.snow.season.length()}},
#' and \code{\link{plt.snow.days.comparison()}}. Data used to produce the plots
#' are saved as RDS files.
#'
#' @param data.model.file Dataset containing the model data. Dataset
#' should be a RDS file in a vectdata format (see \code{\link{vectorize.list()}})
#' @param data.training.file Dataset containing the training data. Dataset
#' should be a RDS file in a vectdata format (see \code{\link{vectorize.list()}})
#' @param data.corrected.file Dataset containing the corrected data. Dataset
#' should be a RDS file in a vectdata format (see \code{\link{vectorize.list()}})
#' @param out.plot.dir Directory where the plots and RDS data should be saved
#' @param swe.threshold The threshold value to define snow day in order to
#' compute the length of the snow season
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
assessment.plot <- function(data.training.file,
                            data.model.file,
                            data.corrected.file,
                            out.plot.dir,
                            swe.threshold,
                            dist.threshold,
                            ma.value = 7,
                            ncores = 1) {

  start.time <- Sys.time()

  dir.create(out.plot.dir, showWarnings = FALSE, recursive = TRUE)

  print("[I] Reading training data")
  data.training<- readRDS(data.training.file)
  print("[I] Reading model data")
  data.model <- readRDS(data.model.file)
  print("[I] Reading corrected data")
  data.corrected <- readRDS(data.corrected.file)

  print(paste("[I] Data read in",
              as.numeric(round(difftime(Sys.time(),start.time, units = "mins"),2)),
              "minutes"))

  print("[I] Plotting seasonnal details")
  start.time.2=Sys.time()
  seasonnal.summary.tmp <- plt.seasonnal.summary.details(
                              data.training = data.training,
                              data.model = data.model,
                              data.corrected = data.corrected,
                              out.plot.dir = out.plot.dir)
  print(paste("[I] Seasonnal details plots produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  pdf(paste0(out.plot.dir,"/summary_calib_valid.pdf"),width=14,height=20)
  layout(matrix(c(1,2,3,4), ncol = 1, byrow = TRUE))
  par(mar=c(2,3,2,1), mgp=c(1.7,0.5,0))

  print("[I] Plotting seasonnal summary")
  start.time.2=Sys.time()
  seasonnal.summary <- plt.seasonnal.summary(seasonnal.summary = seasonnal.summary.tmp)
  print(paste("[I] Seasonnal summary plots produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  print("[I] Plotting snow season length")
  start.time.2=Sys.time()

  snow.season <- plt.snow.season.length(data.training, data.model,
                                        data.corrected,
                                        ncores, swe.threshold, ma.value)
  print(paste("[I] Snow season length plot produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  print("[I] Plotting snow days comparison")
  start.time.2=Sys.time()
  par(mar=c(2,3,2,1), mgp=c(1.7,0.5,0))
  snow.days <- plt.snow.days.comparison(data.training, data.model,
                                        data.corrected, ncores, swe.threshold)
  print(paste("[I] Snow days comparison plot produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  dev.off()


  start.time.2=Sys.time()
  print("[I] Computing seasonal snow days comparison")
  snow.days.season <- snow.days.comparison.seasonnal(data.training, data.model,
                                                     data.corrected, ncores, swe.threshold)
  print(paste("[I] Seasonal snow days comparison computed in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  print("[I] Plotting snow season maps")
  start.time.2=Sys.time()
  plt.snow.season.maps(snow.season.training = snow.season$data.training,
                       snow.season.model = snow.season$data.model,
                       snow.season.corrected = snow.season$data.corrected,
                       out.plot.dir = out.plot.dir)
  print(paste("[I] Snow season maps produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))


  print("[I] Plotting snow days maps")
  start.time.2=Sys.time()
  plt.snow.days.maps(data.model = snow.days$data.model,
                     data.corrected = snow.days$data.corrected,
                     data.model.seasons = snow.days.season$season.model,
                     data.corrected.seasons = snow.days.season$season.corrected,
                     out.plot.dir = out.plot.dir)
  print(paste("[I] Snow snow days maps produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  print("[I] Plotting spatial correlation")
  start.time.2=Sys.time()
  spatial.correlation <- plt.spatial.correlation(data.training = data.training,
                                                 data.model = data.model,
                                                 data.corrected = data.corrected,
                                                 out.plot.dir = out.plot.dir, ncores = ncores,
                                                 dist.threshold = dist.threshold)
  print(paste("[I] Snow spatial correlation plot produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  print("[I] Error pattern computation")
  start.time.2=Sys.time()
  plt.error.pattern(data.training = data.training,
                    data.model = data.model,
                    data.corrected = data.corrected,
                    out.plot.dir = out.plot.dir,
                    ncores = ncores)
  print(paste("[I] Error pattern plot produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  print(paste("[I] Plots produced in",
              as.numeric(round(difftime(Sys.time(),start.time, units = "mins"),2)),
              "minutes"))

  #Clean maps data from output
  snow.season$data.training <- NULL
  snow.season$data.model <- NULL
  snow.season$data.corrected <- NULL

  snow.days$data.model <- NULL
  snow.days$data.corrected <- NULL

  snow.days.season$season.model <- NULL
  snow.days.season$season.corrected <- NULL


  out <- list("seasonnal.summary" = seasonnal.summary, "snow.season" = snow.season,
              "snow.days" = snow.days, "snow.days.season" = snow.days.season,
              "spatial.correlation" = spatial.correlation)
  saveRDS(out,paste0(out.plot.dir,"/summary_calib.RDS"))

  print("[I] Cleaning memory")
  rm(list=ls())
  invisible(gc())
}


assessment.plot.calib.internal <- function (data.training.calib,
                                            data.model.calib,
                                            data.corrected.calib,
                                            data.training.valid,
                                            data.model.valid,
                                            data.corrected.valid,
                                            out.plot.dir,
                                            swe.threshold,
                                            dist.threshold,
                                            ma.value,
                                            ncores)
{

  dir.create(out.plot.dir, showWarnings = FALSE, recursive = TRUE)

  print("[I] Plotting seasonnal details")
  start.time.2=Sys.time()
  seasonnal.summary.tmp <- plt.seasonnal.summary.details.calib(
                                            data.training.calib = data.training.calib,
                                            data.model.calib = data.model.calib,
                                            data.corrected.calib = data.corrected.calib,
                                            data.training.valid = data.training.valid,
                                            data.model.valid = data.model.valid,
                                            data.corrected.valid = data.corrected.valid,
                                            out.plot.dir = out.plot.dir)

  print(paste("[I] Seasonnal details plots produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  pdf(paste0(out.plot.dir,"/summary_calib_valid.pdf"),width=14,height=20)
  layout(matrix(c(1,2,1,2,3,5,4,6,7,7,8,8), ncol = 2, byrow = TRUE))
  par(mar=c(2,3,2,1), mgp=c(1.7,0.5,0))

  print("[I] Plotting seasonnal summary")
  start.time.2=Sys.time()
  seasonnal.summary <- plt.seasonnal.summary.calib(seasonnal.summary = seasonnal.summary.tmp)
  print(paste("[I] Seasonnal summary plots produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  print("[I] Plotting snow season length")
  start.time.2=Sys.time()

  snow.season <- plt.snow.season.length.calib(data.training.calib, data.model.calib,
                                              data.corrected.calib, data.training.valid,
                                              data.model.valid, data.corrected.valid,
                                              ncores, swe.threshold, ma.value)
  print(paste("[I] Snow season length plot produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  print("[I] Plotting snow days comparison")
  start.time.2=Sys.time()
  par(mar=c(2,3,2,1), mgp=c(1.7,0.5,0))
  snow.days <- plt.snow.days.comparison.calib(data.training.calib, data.model.calib,
                                              data.corrected.calib, data.training.valid,
                                              data.model.valid, data.corrected.valid,
                                              ncores, swe.threshold)
  print(paste("[I] Snow days comparison plot produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  dev.off()

  start.time.2=Sys.time()
  print("[I] Computing seasonal snow days comparison")
  snow.days.season <- snow.days.comparison.seasonnal.calib(data.training.calib, data.model.calib,
                                                           data.corrected.calib, data.training.valid,
                                                           data.model.valid, data.corrected.valid,
                                                           ncores, swe.threshold)
  print(paste("[I] Seasonal snow days comparison computed in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  print("[I] Plotting snow season maps")
  start.time.2=Sys.time()
  plt.snow.season.maps.calib(snow.season.training.calib = snow.season$data.training.calib,
                             snow.season.model.calib = snow.season$data.model.calib,
                             snow.season.corrected.calib = snow.season$data.corrected.calib,
                             snow.season.training.valid = snow.season$data.training.valid,
                             snow.season.model.valid = snow.season$data.model.valid,
                             snow.season.corrected.valid = snow.season$data.corrected.valid,
                             out.plot.dir = out.plot.dir)
  print(paste("[I] Snow season maps produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))


  print("[I] Plotting snow days maps")
  start.time.2=Sys.time()
  plt.snow.days.maps.calib(data.model.calib = snow.days$data.model.calib,
                           data.corrected.calib = snow.days$data.corrected.calib,
                           data.model.valid = snow.days$data.model.valid,
                           data.corrected.valid = snow.days$data.corrected.valid,
                           data.model.calib.seasons = snow.days.season$season.model.calib,
                           data.corrected.calib.seasons = snow.days.season$season.corrected.calib,
                           data.model.valid.seasons = snow.days.season$season.model.valid,
                           data.corrected.valid.seasons = snow.days.season$season.corrected.valid,
                           out.plot.dir = out.plot.dir)
  print(paste("[I] Snow snow days maps produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  print("[I] Plotting spatial correlation")
  start.time.2=Sys.time()
  spatial.correlation <- plt.spatial.correlation.calib(data.training.calib = data.training.calib,
                                                       data.model.calib = data.model.calib,
                                                       data.corrected.calib = data.corrected.calib,
                                                       data.training.valid = data.training.valid,
                                                       data.model.valid = data.model.valid,
                                                       data.corrected.valid = data.corrected.valid,
                                                       out.plot.dir = out.plot.dir, ncores = ncores,
                                                       dist.threshold = dist.threshold)
  print(paste("[I] Snow spatial correlation plot produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  print("[I] Plotting error pattern")
  start.time.2=Sys.time()
  plt.error.pattern.calib(data.training.calib = data.training.calib,
                          data.model.calib = data.model.calib,
                          data.corrected.calib = data.corrected.calib,
                          data.training.valid = data.training.valid,
                          data.model.valid = data.model.valid,
                          data.corrected.valid = data.corrected.valid,
                          out.plot.dir = out.plot.dir,
                          ncores = ncores)
  print(paste("[I] error pattern plot produced in",
              as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
              "minutes"))

  #Clean maps data from output
  snow.season$data.training.calib <- NULL
  snow.season$data.model.calib <- NULL
  snow.season$data.corrected.calib <- NULL
  snow.season$data.training.valid <- NULL
  snow.season$data.model.valid <- NULL
  snow.season$data.corrected.valid <- NULL

  snow.days$data.model.calib <- NULL
  snow.days$data.corrected.calib<- NULL
  snow.days$data.model.valid <- NULL
  snow.days$data.corrected.valid <- NULL

  snow.days.season$season.model.calib  <- NULL
  snow.days.season$season.corrected.calib  <- NULL
  snow.days.season$season.model.valid  <- NULL
  snow.days.season$season.corrected.valid  <- NULL

  #Print output and clean memory
  out <- list("seasonnal.summary" = seasonnal.summary, "snow.season" = snow.season,
              "snow.days" = snow.days, "snow.days.season" = snow.days.season,
              "spatial.correlation" = spatial.correlation)

  out
  saveRDS(out,paste0(out.plot.dir,"/summary_calib_valid.RDS"))

}


#' Produce plots to assess the downscaling quality
#'
#' \code{assessment.plot} produces two files (<out.plot.dir>/summary.pdf and
#' <out.plot.dir>/seasonnal_summary.pdf) to asses the quality of the downscaling.
#' The plots are produced using: \code{\link{plt.seasonnal.summary.details()}},
#' \code{\link{plt.seasonnal.summary()}}, \code{\link{plt.snow.season.length()}},
#' and \code{\link{plt.snow.days.comparison()}}. Data used to produce the plots
#' are saved as RDS files.
#'
#' @param data.model.file Dataset containing the model data. Dataset
#' should be a RDS file in a vectdata format (see \code{\link{vectorize.list()}})
#' @param data.training.file Dataset containing the training data. Dataset
#' should be a RDS file in a vectdata format (see \code{\link{vectorize.list()}})
#' @param data.corrected.file Dataset containing the corrected data. Dataset
#' should be a RDS file in a vectdata format (see \code{\link{vectorize.list()}})
#' @param out.plot.dir Directory where the plots and RDS data should be saved
#' @param swe.threshold The threshold value to define snow day in order to
#' compute the length of the snow season
#' @param ncores Number of cores to use for parallel computation, using ncores =
#' 1 defaults to serial computing
#'
#' @author Adrien Michel, 2021 (WSL/SLF)
#' @export
assessment.plot.calib <- function(data.training.calib.file,
                                  data.model.calib.file,
                                  data.corrected.calib.file,
                                  data.training.valid.file,
                                  data.model.valid.file,
                                  data.corrected.valid.file,
                                  out.plot.dir,
                                  swe.threshold,
                                  dist.threshold,
                                  ma.value = 7,
                                  ncores = 1,
                                  mask.files = NULL) {

  dir.create(out.plot.dir, showWarnings = FALSE, recursive = TRUE)

  start.time <- Sys.time()

  print("[I] Reading training data calibration")
  data.training.calib <- readRDS(data.training.calib.file)
  print("[I] Reading model data calibration")
  data.model.calib <- readRDS(data.model.calib.file)
  print("[I] Reading corrected data calibration")
  data.corrected.calib <- readRDS(data.corrected.calib.file)

  print("[I] Reading training data validation")
  data.training.valid <- readRDS(data.training.valid.file)
  print("[I] Reading model data validation")
  data.model.valid <- readRDS(data.model.valid.file)
  print("[I] Reading corrected data validation")
  data.corrected.valid<- readRDS(data.corrected.valid.file)

  print(paste("[I] Data read in",
              as.numeric(round(difftime(Sys.time(),start.time, units = "mins"),2)),
              "minutes"))

  if(is.null(mask.files)){

    assessment.plot.calib.internal(data.training.calib = data.training.calib,
                                   data.model.calib = data.model.calib,
                                   data.corrected.calib = data.corrected.calib,
                                   data.training.valid = data.training.valid,
                                   data.model.valid = data.model.valid,
                                   data.corrected.valid = data.corrected.valid,
                                   out.plot.dir = out.plot.dir,
                                   swe.threshold = swe.threshold,
                                   dist.threshold = dist.threshold,
                                   ma.value = ma.value,
                                   ncores = ncores)

  } else{
    for(mask.file in mask.files) {

      start.time.2 <- Sys.time()
      print(paste("[I] Start computing masks for:", mask.file))

      mask <- read.mask.tif(mask.file)
      mask.name <- tools::file_path_sans_ext(basename(mask.file))
      data.training.calib.mask <- mask.vectdata(data.training.calib, mask)
      data.model.calib.mask <- mask.vectdata(data.model.calib, mask)
      data.corrected.calib.mask <- mask.vectdata(data.corrected.calib, mask)
      data.training.valid.mask <- mask.vectdata(data.training.valid, mask)
      data.model.valid.mask <- mask.vectdata(data.model.valid, mask)
      data.corrected.valid.mask <- mask.vectdata(data.corrected.valid, mask)
      out.plot.dir.mask <- paste0(out.plot.dir,"/",mask.name)

      print(paste("[I] Masking prefromed in",
                  as.numeric(round(difftime(Sys.time(),start.time.2, units = "mins"),2)),
                  "minutes"))

      assessment.plot.calib.internal(data.training.calib = data.training.calib.mask,
                                     data.model.calib = data.model.calib.mask,
                                     data.corrected.calib = data.corrected.calib.mask,
                                     data.training.valid = data.training.valid.mask,
                                     data.model.valid = data.model.valid.mask,
                                     data.corrected.valid = data.corrected.valid.mask,
                                     out.plot.dir = out.plot.dir.mask,
                                     swe.threshold = swe.threshold,
                                     dist.threshold = dist.threshold,
                                     ma.value = ma.value,
                                     ncores = ncores)
    }
  }

  print(paste("[I] Analyse produced in",
              as.numeric(round(difftime(Sys.time(),start.time, units = "mins"),2)),
              "minutes"))

  print("[I] Cleaning memory")
  rm(list=ls())
  invisible(gc())
}

